//------------------------------------------------------------------------
//
// OpenDCC - OpenDecoder - SignalDecoder
//
// Copyright (c) 2006 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      SignalSingleDecoder.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-09-20 V0.01 copied from SignalDecoder
//                             we run on slithly different hardware
//                             - no Jumper JP1
//                             - no ProgLED
//                             - Fixed output to 4 LED pairs
//            2008-02-06 V0.02 changed preamble to 10 bits
//            2013-01-25 V0.03 changed for LED-Dekoder Hardware, only one or two output pairs.
//
//------------------------------------------------------------------------
//
// purpose:   flexible, lowcost decoder for dcc
//            use as decoder for signals with dimming engine
//
// content:   A DCC-Decoder for ATtiny2313 and other AVR
//
//             1. Defines and variable definitions
//             2. DCC receive routine
//             3. Timing values for the timing engine
//             4. Timing engine (pulse durations, blinking)
//             5. MAIN: analyze command, call the action, do programming
//             6. Test and Simulation
//
// issues:    up to now - only linear dimming
//            can't use gcc's srand() and rand() functions - approx. 400 bytes
//            of code! prbs-code from LightDecoder
//            
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#define SIMULATION          0    // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine
                                 // 3: test action of running light

#define PRBS                0

#define NUM_OF_SIGNALS      1    // 1 oder 2


//---------------------------------------------------------------------
// Timing Definitions:
//
#ifndef F_CPU
   // prevent compiler error by supplying a default 
   # warning "F_CPU not defined for <SignalSingleDecoder.c>", set default to 8MHz 
   # define F_CPU 8000000UL
   // if changed: check every place where it is used
   // (possible range underflow/overflow in preprocessor!)
#endif
#if (F_CPU == 10000000UL)
    const unsigned char speed[] PROGMEM = {"..10MHz.."};
#endif
#if (F_CPU == 8000000UL)
    const unsigned char speed[] PROGMEM = {"..8MHz.."};
#endif

/// This are DCC timing definitions from NMRA
#define PERIOD_1   116L          // 116us for DCC 1 pulse - do not change
#define PERIOD_0   232L          // 232us for DCC 0 pulse - do not change


// Definitions for the DIMM-Engine
// see below on information about the DIMM-Engine

#define PWM_TICK_PERIOD     300L        // 300us tick for PWM Engine
#define PWM_STEPS           60L
#define PWM_PERIOD          (PWM_TICK_PERIOD * PWM_STEPS)    // 18ms = 300 * 60L
#define DIMM_UP_DELAY       400000L     // Dimm up with 400ms delay 
#define DIMM_DOWN_SPEED     4L          // Dimm down slow 
#define DIMM_UP_SPEED       5L          // Dimm up a little faster
#define DIMM_DOWN_SAME_BULB 1           // 1: wenn im neuen Begriff die lampe 
                                        //    wieder leuchtet, zuerst abdimmen,
                                        //    dann wieder aufdimmen.
                                        // 0: leuchtet durchgehend  



//---------------------------------------------------------------------------
// PORT Definitions:
//
// PORTD:
#define PROGTASTER    0     // input, internal pullup
#define DCCIN         2     // must be located on INT0
#define LED_RED1      5     // out
#define LED_GREEN1    4     // out

// PORTB:
#define LED_RED2      0     // out
#define LED_GREEN2    1     // out

#define PROG_PRESSED   (!(PIND & (1<<PROGTASTER)))
#define JUMPER_FITTED  (!(PIND & (1<<JUMPER)))
#define LED_RED1_OFF   PORTD |= (1<<LED_RED1)
#define LED_RED1_ON    PORTD &= ~(1<<LED_RED1)

//----------------------------------------------------------------------------
// Global Data

// unsigned int  MyAdr;               // my DCC address
unsigned char MyOpMode;            // mode of operation

volatile unsigned char MyDelay;    // is decremented by ISR

unsigned int  ReceivedAdr;         // last received
unsigned char ReceivedCommand;


// 
unsigned char hauptsignal_state;    // fr Dunkeltastung
unsigned char vorsignal_state;

//-----------------------------------------------------------------------------
// data in EEPROM:
//
unsigned char EE_empty     EEMEM = 0xff; // keep free
unsigned char EE_myAdrL    EEMEM = 0x00; // Decoder Adresse low
unsigned char EE_myAdrH    EEMEM = 0x80; // Decoder Adresse high (MSB set to mark we are unprogrammmed)
unsigned char EE_myOpMode  EEMEM = 0;    // Pulse oder Dauer
unsigned char EE_LastState EEMEM = 0;    // aktueller Portzustand

//------------------------------------------------------------------------
// OpMode: (loaded by coil during programming)
// 0: Vier 2-begriffige Signale
// 1: Zwei 3-begriffige Signale
// 2: Einfahrsignal + Vorsignal getrennt 
// 3: Einfahrsignal + Vorsignal mit Dunkeltastung
// 4: zwei Wechselblinker (mit Ansteuerung Stopmagnet)
// 5: tbd. (?? ev. zwei Signale + 2 Schalter)
// 6: reserved 
// 7: Ampel mit Ansteuerung Stopmagnet
//
//---------------------------------------------------------------------------------
// Some tricks to optmize size and speed:
//
// 1. Save global flags (like Communicate or dcc.state) in unused IOs
// 2. Speed up port access by declaring port access as inline code
// 3. Cast logical operation down to char, whenever possible.
//    use assign to local variable and compare this var - smaller size
// 4. Same code in different cases should be identically ordered -
//    gcc reuses this code!
//
// Trick 1: Using IO as variables - what are unused IOs on ATtiny2313:
//   GPIOR0, GPIOR1, GPIOR2 General Purpose I/O Register 
//   UBRRL Baudrate, (USIDR USI Data Register)
//
     #define OPTIMIZE_VARS
//
//   We use:
//      GPIOR0 for general communication between processes
//      GPIOR1 for DCC ISR
//      GPIOR2 for Output State: 1 = ramp towards 1, 0 = ramp towards 0
    
     #ifdef OPTIMIZE_VARS
         #define Communicate    GPIOR0
         #define Recstate       GPIOR1
         #define CurrentTarget  GPIOR2   
         unsigned char CurrentAspect;
     #else
         unsigned char Communicate;
         unsigned char Recstate;
         unsigned char CurrentTarget;
         unsigned char CurrentAspect;
                 
     #endif


// Trick 2: Speed up port access by declaring port access as inline code
//          This is done with a subroutine; if gcc runs with -Os, this results in
//          single cbi and sbi statements!

// not used here //
static inline void output(unsigned char port, unsigned char state) 
       __attribute__((always_inline));

void output(unsigned char port, unsigned char state)
  {
    if (state == 0)
      {                              
        PORTB &= (~(1<<port));
      }
    else
      {
        PORTB |= (1<<port);
      }
  }
                      
static inline void set_target(unsigned char port, unsigned char state) 
       __attribute__((always_inline));

void set_target(unsigned char port, unsigned char state)
  {
    if (state == 0)
      {                              
        CurrentTarget &= (~(1<<port));
        CurrentAspect &= (~(1<<port));

      }
    else
      {
        CurrentTarget |= (1<<port);
        CurrentAspect |= (1<<port);
      }
  }
                      

//----------------------------------------------------------------------------------
// Definitions for inter process communication


#define C_Received        0    // a new DCC was received - issued by ISR(Timer1)
                               //                          cleared by main
#define C_DoSave          1    // a new PORT state should be saved
                               //                        - issued by action
                               //                          cleared by main
#define C_Dimmstep        2    // perform recalculation of dimm_val
                               //                        - issued by do_dimm every 10ms
                               //                          cleared by main (dimmer)
    




void init_main(void)
  {
    DDRB  = (1<<LED_RED2)  
          | (1<<LED_GREEN2) ; 

    PORTB  = (0<<LED_RED2)  
           | (0<<LED_GREEN2) ;    // both on
           
    DDRD  = (0<<PROGTASTER)     // in - set to 0
          | (0<<DCCIN)  
          | (1<<LED_RED1)  
          | (1<<LED_GREEN1) ; 
    PORTD = (1<<PROGTASTER)     // pullup - set to 1
          | (0<<DCCIN) 
          | (0<<LED_RED1)  
          | (0<<LED_GREEN1) ;    // both on

    // Init Timer1

    TCCR1B = (0 << ICNC1) 
           | (0 << ICES1) 
           | (0 << WGM13) 
           | (1 << WGM12)       // Mode 4: CTC with OCRA1
           | (0 << CS12)        // clk stopped
           | (0 << CS11) 
           | (0 << CS10);

    // set Timer1 Compare to 3/4 of period of a one -> this is 116*0,75=87us
    
    OCR1A = F_CPU * PERIOD_1 * 3 / 4 / 1000000L;  

    // Init Timer0 as CTC 
    // check PWM_TICK_PERIOD and F_CPU
    #define TIMER0_CLKDIV 64              // possible values: 1, 8, 64, 256, 1024

    #if (F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER0_CLKDIV) > 255L
      #warning: overflow in OCR0A - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a larger clkdiv
    #endif    
    #if (F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER0_CLKDIV) < 30L
      #warning: resolution accuracy in OCR0A too low - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a smaller clkdiv
    #endif    


    OCR0A = F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER0_CLKDIV ;  
    TCCR0A = (0 << COM0A1)      // compare match A
           | (0 << COM0A0) 
           | (0 << COM0B1)      // compare match B
           | (0 << COM0B0) 
           | 0                  // reserved
           | 0                  // reserved
           | (1 << WGM01)  
           | (0 << WGM00);      // Timer0 Mode 2 = CTC - Int on compare A
    TCCR0B = (0 << FOC0A) 
           | (0 << FOC0B) 
           | (0 << WGM02) 
    #if   TIMER0_CLKDIV == 1
           | (0 << CS02) | (0 << CS01)  | (1 << CS00);       // CS[2:0]=001 clkdiv 1
    #elif TIMER0_CLKDIV == 8
           | (0 << CS02) | (1 << CS01)  | (0 << CS00);       // CS[2:0]=010 clkdiv 8
    #elif TIMER0_CLKDIV == 64
           | (0 << CS02) | (1 << CS01)  | (1 << CS00);       // CS[2:0]=011 clkdiv 64 
    #elif TIMER0_CLKDIV == 256
           | (1 << CS02) | (0 << CS01)  | (0 << CS00);       // CS[2:0]=100 clkdiv 256
    #elif TIMER0_CLKDIV == 1024
           | (1 << CS02) | (0 << CS01)  | (1 << CS00);       // CS[2:0]=101 clkdiv 1024   
    #else
     #warning: TIMER0_CLKDIV is void
    #endif


    TIMSK = (0<<TOIE1)       // Timer1 Overflow
          | (1<<OCIE1A)      // Timer1 Compare A
          | (0<<OCIE1B)      // Timer1 Compare B
          | 0                // reserved
          | (0<<ICIE1)       // Timer1 Input Capture
          | (0<<OCIE0B)      // Timer0 Compare B
          | (0<<TOIE0)       // Timer0 Overflow
          | (1<<OCIE0A);     // Timer0 Compare A


    // Init Interrupt 0

    GIMSK = (0<<INT1)        // Int1 is msb
          | (1<<INT0)        // Enable INT0
          | (0<<PCIE); 


    MCUCR = 0x03;  //           ;The rising edge of INT0 generates an interrupt request.

          
  }






//==============================================================================
//
// Section 2
//
// DCC Receive Routine
//
// Howto:    uses two interrupt: a rising edge in DCC polarity triggers INT0;
//           in INT0, Timer1 with a delay of 87us is started.
//           On Timer1 Compare Match the level of DCC is evaluated and
//           parsed.
//
//                           |<-----116us----->|
//
//           DCC 1: _________XXXXXXXXX_________XXXXXXXXX_________
//                           ^-INT0
//                           |----87us--->|
//                                        ^-INT1: reads zero
//
//           DCC 0: _________XXXXXXXXXXXXXXXXXX__________________
//                           ^-INT0
//                           |----------->|
//                                        ^-INT1: reads one
//           
// Result:   1. The received message is stored in "message" and "message_size"
//           2. The flag C_Received is set.
//
#define MAX_MESSAGE 15    // including XOR-Byte
volatile unsigned char message[MAX_MESSAGE];
volatile unsigned char message_size;

struct
    {
        unsigned char state;                    // current state
        unsigned char bitcount;                 // current bit
        unsigned char bytecount;                // current byte
        unsigned char accubyte;                 // actual check
    } dccrec;

// some states:
#define RECSTAT_WF_PREAMBLE  0
#define RECSTAT_WF_LEAD0     1
#define RECSTAT_WF_BYTE      2
#define RECSTAT_WF_TRAILER   3

#define RECSTAT_DCC          7   


// ISR(INT0) loads only a register and stores this register to IO.
// this influences no status flags in SREG.
// therefore we define a naked version of the ISR with
// no compiler overhead.

#define ISR_INT0_OPTIMIZED

#ifdef ISR_INT0_OPTIMIZED
    #ifdef ISR_NAKED
    #undef ISR_NAKED
    #endif
    #define ISR_NAKED(vector) \
        void vector (void) __attribute__ ((signal, naked)); \
        void vector (void)

    ISR_NAKED(INT0_vect) 
      {
         __asm__ __volatile 
          (
            "push r16"  "\n\t"
            "ldi r16, %1"  "\n\t"
            "out %0, r16" "\n\t"
            "pop r16"  "\n\t"
         :                         // no output section
         : "M" (_SFR_IO_ADDR (TCCR1B)),
           "M" ((0 << ICNC1)       // start timer1
              | (0 << ICES1) 
              | (0 << WGM13) 
              | (1 << WGM12)       // Mode 4: CTC
              | (0 << CS12)        // clk 1:1
              | (0 << CS11) 
              | (1 << CS10))
          );
        asm volatile ( "reti" ); 
      }
#else
    ISR(INT0_vect) 
      {
        TCCR1B = (0 << ICNC1)       // start timer1
               | (0 << ICES1) 
               | (0 << WGM13) 
               | (1 << WGM12)       // Mode 4: CTC
               | (0 << CS12)        // clk 1:1
               | (0 << CS11) 
               | (1 << CS10); 
      }

#endif

const unsigned char copy[] PROGMEM = {"..SignalSignleDecoder V0.03 (c) Kufer 2006-2013.."};

ISR(TIMER1_COMPA_vect)
  {
    #define mydcc (Recstate & (1<<RECSTAT_DCC))

    // read asap to keep timing!
    if (PIND & (1<<DCCIN)) Recstate &= ~(1<<RECSTAT_DCC);  // if high -> mydcc=0
    else                   Recstate |= 1<<RECSTAT_DCC;    

    //sei(); !!!

    TCCR1B = (0 << ICNC1)       
           | (0 << ICES1) 
           | (0 << WGM13) 
           | (1 << WGM12) 
           | (0 << CS12)        // clk stopped
           | (0 << CS11) 
           | (0 << CS10);

    TCNT1 = 0;                  // clear Counter

    dccrec.bitcount++;

    if (Recstate & (1<<RECSTAT_WF_PREAMBLE))            // wait for preamble
      {                                       
        if (mydcc)
          {
            if (dccrec.bitcount >= 10) 
              {
                Recstate = 1<<RECSTAT_WF_LEAD0;            
              }
          }
        else
          {
            dccrec.bitcount=0;
          }
      }
    else if (Recstate & (1<<RECSTAT_WF_LEAD0))          // wait for leading 0
      {
        if (mydcc)
          {                                             // still 1, wait again
          }
        else
          {
            dccrec.bytecount=0;
            Recstate = 1<<RECSTAT_WF_BYTE;
            dccrec.bitcount=0;
            dccrec.accubyte=0;
          }
      }
    else if (Recstate & (1<<RECSTAT_WF_BYTE))           // wait for byte
      {
        unsigned char my_accubyte;
        my_accubyte = dccrec.accubyte << 1;
        if (mydcc)
          {
            my_accubyte |= 1;
          }
        dccrec.accubyte = my_accubyte;
        
        /* dccrec.accubyte = dccrec.accubyte << 1;
        if (mydcc)
          {
            dccrec.accubyte |= 1;
          }
         */
        if (dccrec.bitcount==8)
          {
            if (dccrec.bytecount == MAX_MESSAGE)        // too many bytes
              {                                         // ignore message
                Recstate = 1<<RECSTAT_WF_PREAMBLE;   
              }
            else
              {
                message[dccrec.bytecount++] = dccrec.accubyte;
                Recstate = 1<<RECSTAT_WF_TRAILER; 
              }
          }
      }
    else if (Recstate & (1<<RECSTAT_WF_TRAILER))        // wait for 0 (next byte) 
      {                                                 // or 1 (eof message)
        if (mydcc)
          {  // trailing "1" received
            Recstate = 1<<RECSTAT_WF_PREAMBLE;
            dccrec.bitcount=1;
            message_size = dccrec.bytecount;
            Communicate |= (1<<C_Received);
          }
        else
          {
            Recstate = 1<<RECSTAT_WF_BYTE;
            dccrec.bitcount=0;
            dccrec.accubyte=0;
          }
      }
    else
      {
        Recstate = 1<<RECSTAT_WF_PREAMBLE;
      }
  }

#if (SIMULATION == 1)

unsigned char dccbit;
void simulat_receive(void);

void dcc_receive(void)
  {
    #define mydcc dccbit

    dccrec.bitcount++;

    if (Recstate & (1<<RECSTAT_WF_PREAMBLE))            // wait for preamble
      {                                       
        if (mydcc)
          {
            if (dccrec.bitcount >= 10) 
              {
                Recstate = 1<<RECSTAT_WF_LEAD0;            
              }
          }
        else
          {
            dccrec.bitcount=0;
          }
      }
    else if (Recstate & (1<<RECSTAT_WF_LEAD0))          // wait for leading 0
      {
        if (mydcc)
          {                                             // still 1, wait again
          }
        else
          {
            dccrec.bytecount=0;
            Recstate = 1<<RECSTAT_WF_BYTE;
            dccrec.bitcount=0;
            dccrec.accubyte=0;
          }
      }
    else if (Recstate & (1<<RECSTAT_WF_BYTE))           // wait for byte
      {
        dccrec.accubyte = dccrec.accubyte << 1;
        if (mydcc)
          {
            dccrec.accubyte |= 1;
          }
        if (dccrec.bitcount==8)
          {
            if (dccrec.bytecount == MAX_MESSAGE)        // too many bytes
              {                                         // ignore message
                Recstate = 1<<RECSTAT_WF_PREAMBLE;   
              }
            else
              {
                message[dccrec.bytecount++] = dccrec.accubyte;
                Recstate = 1<<RECSTAT_WF_TRAILER; 
              }
          }
      }
    else if (Recstate & (1<<RECSTAT_WF_TRAILER))        // wait for 0 (next byte) 
      {                                                 // or 1 (eof message)
        if (mydcc)
          {  // trailing "1" received
            Recstate = 1<<RECSTAT_WF_PREAMBLE;
            dccrec.bitcount=1;
            message_size = dccrec.bytecount;
            Communicate |= (1<<C_Received);
          }
        else
          {
            Recstate = 1<<RECSTAT_WF_BYTE;
            dccrec.bitcount=0;
            dccrec.accubyte=0;
          }
      }
    else
      {
        Recstate = 1<<RECSTAT_WF_PREAMBLE;
      }
  }

#endif //  (SIMULATION == 1)


#define DIMM_ENGINE
#ifdef DIMM_ENGINE

//==============================================================================
//
// Section 3
//
// Lichtsteuerung als PWM-Dimmer
//
//             |                        |                          |
// Output:     |XXXXXXXXXX______________|XXXXXXXXXX________________|
//             |<-------->              |<------------------------>|
//             | dimm_val               |        PWM_PERIOD        |
//         ----|------------------------|--------------------------|---> Time
//
//
// 1. Einstellen des aktuellen Helligkeitswertes (do_dimm)
//
//    Es gibt 60 Helligkeitstufen.
//    Alle 300us erfolgt ein Interrupt, dieser schaltet den Dimmer um eine
//    Stufe weiter, nach 60 Stufen wird wieder von vorne begonnen.
//    Stufe MIN:  alle Ports mit einem dimm_val > DIMM_MIN werden eingeschaltet.
//    Stufe x:    Ein Port, dessen dimm_val kleiner x ist, wird abgeschaltet.
//    Stufe MAX:  Restart und Meldung an den DIMMER (C_Dimmstep)
//
//    Folge: Alle Ports mit einem dimm_val kleiner DIMM_RANGE_MIN sind
//    dauerhaft aus, alle Ports mit einem dimm_val grer DIMM_RANGE_MAX
//    sind dauerhaft ein.
//
//    Da der PWM 60 Stufen hat und alle 300us ein Int erfolgt, wird dieser
//    Durchlauf alle 18ms durchgefhrt. Dies entspricht einer Refreshrate von
//    55Hz.
//
//
// 2. Langsame Vernderung der Helligkeit
//
//    Nach einem Zyklus des PWM wird vom Hauptprogramm der neue aktuelle
//    dimm_val ausgerechnet. Hierzu wird vom aktuellen Wert mit einem
//    Schritt "delta" nach oben oder unten gerechnet, bis der neue Zielwert
//    erreicht ist.
//
//    Die Zykluszeit der PWM ist 18ms, somit wird je nach "delta" folgende
//    Dimmzeit erreicht:
//
//      delta    |    Dimmzeit
//    -----------------------------
//        1      |    1080ms
//        2      |     540ms
//        3      |     360ms
//        4      |     270ms
//        5      |     216ms
//        6      |     180ms
//      100      |    sofort
//
//    Der neue Zielwert wird in light_val hinterlegt. 
//
//    Wenn man von einem Wert kleiner DIMM_RANGE_MIN startet, dann wird
//    der Port erst mit Verzgerung aufgedimmt, weil zuerst der Bereich
//    bis DIMM_RANGE_MIN "aufgedimmt" wird.
//    Dies wird dazu benutzt, zuerst das alte Signalbild wegzudimmen 
//    und dann das neue Signalbild aufzudimmen.
//
// 3. Signalbilder
// 
//    Signalbilder werden als Bitfeld hinterlegt. Fr jedes Kommando
//    gibt es ein Bitfeld, in dem das neue Signalbild abgelegt ist und eine
//    Gltigkeitsmaske, diese bestimmt, auf welche Dimmwerte das Signalbild
//    wirken soll. Mit diesen beiden Pattern wird "set_new_light_val" 
//    aufgerufen.
//
// 4. Blinken
//    
//    Falls ontime bzw. offtime ungleich 0 sind, wird nach Ablauf der
//    jeweils andere Phasenwert geladen.
//

#define DIMM_RANGE_MIN 100    // aktiver Bereich 100-160
#define DIMM_RANGE_MAX (DIMM_RANGE_MIN+PWM_STEPS+1)



//------------------------------------------- Array for DIMM-Engine
// a) running values
unsigned char cur_dimm;            //  = DIMM_RANGE_MIN;


// b) target values
volatile struct
  {
    unsigned char rest;     	// Zeit bis zum nchsten Wechsel in PWM_PERIOD (18ms)
    unsigned char ontime;   	// Einschaltzeit
    unsigned char offtime;  	// Ausschaltzeit
	unsigned char dimm_val;		// aktueller Istwert
	unsigned char light_A_val;  // aktueller Zielwert in Richtung A
    unsigned char delta_A;      // aktuelles Increment in Richtung A
    unsigned char light_B_val;  // aktueller Zielwert in Richtung B
    unsigned char delta_B;
    unsigned char ledbit;
  } out_pwm[2*NUM_OF_SIGNALS];

#define SIZE_OUT  (sizeof(out_pwm) / sizeof(out_pwm[0]))

#if (1==0)
//---- general
ISR(TIMER0_COMPA_vect)                      // Timer0 Compare Int
  {                                         // macht pwm
    unsigned char port;

    sei();
    cur_dimm++;
    if (cur_dimm == DIMM_RANGE_MAX)
      {
        cur_dimm = DIMM_RANGE_MIN;
        for (port=0; port<SIZE_OUT; port++)
          {                                             // we are active low
            if (out_pwm[port].dimm_val > DIMM_RANGE_MIN)  PORTD &= ~out_pwm[port].ledbit;  // PORTB |= mask;   // Einschalten wenn !0
          }
        Communicate |= (1<<C_Dimmstep);
        if (MyDelay) MyDelay--; 
      }
    else
      {
        for (port=0; port<SIZE_OUT; port++)
          {
            if (cur_dimm >= out_pwm[port].dimm_val) PORTD |= out_pwm[port].ledbit;   //  PORTB &= ~mask; 
          }
      }
  }
#endif

#if (NUM_OF_SIGNALS == 1)
//---- Single Signal
ISR(TIMER0_COMPA_vect)                      // Timer0 Compare Int
  {                                         // macht pwm
    sei();
    cur_dimm++;
    if (cur_dimm == DIMM_RANGE_MAX)
      {
        cur_dimm = DIMM_RANGE_MIN;
        if (out_pwm[0].dimm_val > DIMM_RANGE_MIN)  PORTD &= ~out_pwm[0].ledbit;  // PORTB |= mask;   // Einschalten wenn !0
        if (out_pwm[1].dimm_val > DIMM_RANGE_MIN)  PORTD &= ~out_pwm[1].ledbit;  // we are active low
        
        Communicate |= (1<<C_Dimmstep);
        if (MyDelay) MyDelay--; 
      }
    else
      {
        if (cur_dimm >= out_pwm[0].dimm_val) PORTD |= out_pwm[0].ledbit;   //  PORTB &= ~mask; 
        if (cur_dimm >= out_pwm[1].dimm_val) PORTD |= out_pwm[1].ledbit;   //  PORTB &= ~mask; 
      }
  }
#endif
#if (NUM_OF_SIGNALS == 2)
//---- DUAL
ISR(TIMER0_COMPA_vect)                      // Timer0 Compare Int
  {                                         // macht pwm
    sei();
    cur_dimm++;
    if (cur_dimm == DIMM_RANGE_MAX)
      {
        cur_dimm = DIMM_RANGE_MIN;
        if (out_pwm[0].dimm_val > DIMM_RANGE_MIN)  PORTD &= ~out_pwm[0].ledbit;  // PORTB |= mask;   // Einschalten wenn !0
        if (out_pwm[1].dimm_val > DIMM_RANGE_MIN)  PORTD &= ~out_pwm[1].ledbit;  // we are active low
        if (out_pwm[2].dimm_val > DIMM_RANGE_MIN)  PORTB &= ~out_pwm[2].ledbit;  // PORTB |= mask;   // Einschalten wenn !0
        if (out_pwm[3].dimm_val > DIMM_RANGE_MIN)  PORTB &= ~out_pwm[3].ledbit;  // we are active low
        
        Communicate |= (1<<C_Dimmstep);
        if (MyDelay) MyDelay--; 
      }
    else
      {
        if (cur_dimm >= out_pwm[0].dimm_val) PORTD |= out_pwm[0].ledbit;   //  PORTB &= ~mask; 
        if (cur_dimm >= out_pwm[1].dimm_val) PORTD |= out_pwm[1].ledbit;   //  PORTB &= ~mask; 
        if (cur_dimm >= out_pwm[2].dimm_val) PORTB |= out_pwm[2].ledbit;   //  PORTB &= ~mask; 
        if (cur_dimm >= out_pwm[3].dimm_val) PORTB |= out_pwm[3].ledbit;   //  PORTB &= ~mask; 
      }
  }
#endif

// Diese Funktion setzt die neuen Ziel-Bits fr DIMM
// beim Aufdimmen gibt es immer 400ms Verzgerung, damit ein Abdimmen vorher fertig wird.


void set_new_light_val(unsigned char pattern, unsigned char valid)
  {
    unsigned char port;
    unsigned char mask;
    unsigned char inv_pattern = ~pattern;

    mask = 1;
    for (port=0; port<SIZE_OUT; port++)
      {
        if (pattern & mask & valid)
          {
            // turn on
            #if DIMM_DOWN_SAME_BULB == 1
                out_pwm[port].rest = DIMM_UP_DELAY / PWM_PERIOD;   
                CurrentTarget &= ~mask;
            #else
                if (CurrentTarget & mask)
                  {
                    // already on, nothing to do, keep burning 
                  }
                else
                  {
                    out_pwm[port].rest = DIMM_UP_DELAY / PWM_PERIOD;   
                    CurrentTarget &= ~mask;
                  }
            #endif
            CurrentAspect |= mask; 
          }
        if (inv_pattern & mask & valid)
          {
            // turn off
            out_pwm[port].rest    = 0;
            CurrentTarget &= ~mask;
            CurrentAspect &= ~mask; 
          }
        mask = mask << 1;
      }
  }




//---------------------------------------------------------------------------------
// dimmer()
// this routine is called from main(), if C_Dimmstep is activated  
//

// Timing Engine
//
// Howto:    
// 1. Generelles Timing:
//    Diese Routine wird alle PWM_PERIOD aufgerufen. Es wird folgendes
//    geprft:
//    a) Wenn out_pwm[port].rest gleich 0: dann bleibt dieser Port unverndert.
//    b) out_pwm[port].rest wird decrementiert, wenn es dabei 0 wird, dann
//       wird ein Dimmvorgang in die andere Richtung eingeleitet.
//
// 2. Dimm-bergnge:
//    Je nach aktueller Richtung des Dimmvorgang (CurrentTarget) wird der aktuelle
//    Dimmwert erhht oder erniedrigt (z.Z. linear).
//    Die Dimmrampe ist unabhngig von den Zeiten, die bei ontime bzw. offtime
//    vorgegeben werden.
//    Wenn ein Ausgang schalten soll, dann mu sein Delta sehr gro gewhlt
//    werden!
  


void dimmer(void)
  {
    unsigned char port;
    unsigned char mask;
    unsigned char my_rest;

    mask = 1;
    for (port=0; port<SIZE_OUT; port++)
      {
        my_rest = out_pwm[port].rest;       // use a local variable to force
        if (my_rest !=0)                    // compiler to tiny code
          { 
            if (--my_rest == 0)
              {
                 if (CurrentTarget & mask)
                  { // bit was on
                    my_rest = out_pwm[port].offtime;
                    CurrentTarget &= ~mask;
                  }
                else
                  {
                    my_rest = out_pwm[port].ontime;
                    CurrentTarget |= mask;
                  }
              }
            out_pwm[port].rest = my_rest;
          }


    	my_rest = out_pwm[port].dimm_val;
        if (CurrentTarget & mask)    
          {                                     // we are in phase A -> incr
            if (my_rest < out_pwm[port].light_A_val)  
              {
                // fehlt Sicherung gegen berlauf -> bei Daten aufpassen
                // delta darf nicht zu gro sein
				my_rest += out_pwm[port].delta_A;
              }
          }
        else
          {                                     // we are in phase B -> decr
            if (my_rest > out_pwm[port].light_B_val)
              {
                if (my_rest > out_pwm[port].delta_B)
					 my_rest -= out_pwm[port].delta_B;
                else my_rest = 0;
              }
          }
        out_pwm[port].dimm_val = my_rest;
        mask = mask << 1;                  // do *not* calc mask from port
      }
  }


#endif // DIMM_ENGINE
//==============================================================================
//

//==============================================================================
//
// Section 4
//
// Pattern for different signals
//
// a) Vierfach rot grn; 
//
// Anschlu:
//
//  0baabbccdd     Anschlu:
//    |  ||  |---- Hauptsignal grn
//    |  || |----- Hauptsignal rot

unsigned char EE_Signal_DB2_Pattern[8] EEMEM =
  {
    0b00000010,          // hp0
    0b00000001,          // hp1
    0b00001000,
    0b00000100,
    0b00100000,
    0b00010000,
    0b10000000,
    0b01000000,
  };

unsigned char EE_Signal_DB2_Valid[8] EEMEM =
  {
    0b00000011,
    0b00000011,
    0b00001100,
    0b00001100,
    0b00110000,
    0b00110000,
    0b11000000,
    0b11000000,
  };


#define PROG_BLINK  ( 700000L / PWM_PERIOD)   // Blinken Progmode

void show_progmode_on(void)
  {
    out_pwm[0].dimm_val    = (DIMM_RANGE_MIN+DIMM_RANGE_MAX) / 2;
    out_pwm[1].dimm_val    = (DIMM_RANGE_MIN+DIMM_RANGE_MAX) / 2;
  }

void show_progmode_off(void)
  {
  }



//==============================================================================
//
// Section 5
//
// MAIN: analyze command, call the action, do programming
//



//------------------------------------------------------------------------------
// This Routine is called when myAdr is received

void action(void)
  {
    unsigned char myCommand;
    // unsigned char myTurnout;

    if (ReceivedCommand >= NUM_OF_SIGNALS * 2) return; 
    
    // myTurnout = ReceivedCommand & 0b00001110;    // mask turnout without coil   
    // myCommand = ReceivedCommand & 0b00000111;
    myCommand = ReceivedCommand;     

    cli();                      // block interrupts

    // Vier zweibegriffige Signale
      {
        set_new_light_val(
            eeprom_read_byte(&EE_Signal_DB2_Pattern[myCommand]),
            eeprom_read_byte(&EE_Signal_DB2_Valid[myCommand])    );
        Communicate |= (1<<C_DoSave); 
      }
    sei();
  }

//
//-----------------------------------------------------------------------------
// analyze_message checks the received DCC message
// return 0 if void, 1 if accessory command, 2 if myAdr or myAdr+1;
//
unsigned char analyze_message(void)
  {
    unsigned char i;
    unsigned char myxor = 0;
    unsigned int MyAdr;
    
    for (i=0; i<message_size; i++)
      {
        myxor = myxor ^ message[i];
      }

    if (myxor)
      {
        // checksum error, ignore
        return(0);
      }
    else
      {
        // check, if it is an accessory message (128-192)
        myxor = message[0] & 0b11000000;
        if (myxor == 0b10000000)
          {
            if (message[1] >= 0b10000000)  // MSB in Command byte set
              {
                if (message[1] & (1<<3))   // Bit 3: accessory command + active coil?
                  {
                    ReceivedCommand = message[1] & 0b00000111;
    
                    // take bits 5 4 3 2 1 0 from message[0]
                    // take Bits 6 5 4 from message[1] and invert
    
                    #define OPTCODE1
                    #ifdef OPTCODE1
                        unsigned char temp;
                        myxor = ~message[1] & 0b01110000;
                        myxor = myxor<<1;  // shift as byte
                        temp = message[0] & 0b00111111;
                        ReceivedAdr = (myxor<<1) | temp;
                    #else 
                        ReceivedAdr = (message[0] & 0b00111111)
                                    | ((~message[1] & 0b01110000) << 2);
                    #endif
    
                    MyAdr = (eeprom_read_byte(&EE_myAdrH)  << 8) |
                            (eeprom_read_byte(&EE_myAdrL));
    
                    if (ReceivedAdr == MyAdr) return(2);
                    else if (ReceivedAdr == MyAdr+1)
                      {
                        ReceivedCommand += 8;
                        return(2);
                      } 
                    else return(1);
                  }
              }
          }
      }
    return(0);
  }


//------------------------------------------------------------------------
// This Routine is called when PROG is pressed
//
#define DEBOUNCE  (50000L / PWM_PERIOD)
#if (DEBOUNCE == 0)
 #define DEBOUNCE  1
#endif


void DoProgramming(void)
  {
    unsigned char myCommand;    

    cli();
    MyDelay = DEBOUNCE;
    sei();

    while(MyDelay) ;     // wait until ISR has decremented  MyDelay

    if (PROG_PRESSED)                    // still pressed?
      {
        LED_RED1_ON;
        show_progmode_on();             // both LED half
        Communicate &= ~(1<<C_Received);
        
        while(PROG_PRESSED) ;           // wait for release

        cli();
        MyDelay = DEBOUNCE;
        sei();

        while(MyDelay) ;     // wait until ISR has decremented  MyDelay
        

        while(!PROG_PRESSED)
          {
            if (Communicate & (1<<C_Received))
              {                                     // Message
                Communicate &= ~(1<<C_Received);
                if (analyze_message())              // Accessory
                  {
                    eeprom_write_byte(&EE_myAdrL, (unsigned char) ReceivedAdr);     
                    eeprom_write_byte(&EE_myAdrH, (unsigned char) (ReceivedAdr >> 8));
                    
                    myCommand = ReceivedCommand & 0x06;         // mask turnout, not coil
                    eeprom_write_byte(&EE_myOpMode, myCommand);
                    MyOpMode = myCommand;

                    eeprom_write_byte(&EE_LastState, 0);
                    
                    do {} while (!eeprom_is_ready());    // wait for write to complete
                    
                    LED_RED1_OFF;
                    
                    // we got reprogrammed ->
                    // forget everthing running and restart decoder!                    
                    
                    // cli();
                    
                    // laut diversen Internetseiten sollte folgender Code laufen -
                    // tuts aber nicht, wenn man das Assemblerlistung ansieht.
                    // void (*funcptr)( void ) = 0x0000;    // Set up function pointer
                    // funcptr();                        // Jump to Reset vector 0x0000
                    
                    __asm__ __volatile 
                    (
                       "ldi r30,0"  "\n\t"
                       "ldi r31,0"  "\n\t"
                       "icall" "\n\t"
                     );
                    // return;  
                  }
              }
          }  // while
        LED_RED1_OFF;
        show_progmode_off();        
        cli();
        MyDelay = DEBOUNCE;
        sei();
        while(MyDelay) ;     // wait until ISR has decremented  MyDelay
        while(PROG_PRESSED) ;           // wait for release
      }
    return;   
  }


int main(void)
  {
    unsigned char port;
    init_main();

    // Delta fr Glhlampensimulation vorbelegen

    for (port=0; port<SIZE_OUT; port++)
          {
            out_pwm[port].dimm_val    = DIMM_RANGE_MIN;
            out_pwm[port].delta_A     = DIMM_UP_SPEED; 
            out_pwm[port].light_A_val = DIMM_RANGE_MAX;                
			out_pwm[port].delta_B     = DIMM_DOWN_SPEED;
			out_pwm[port].light_B_val = DIMM_RANGE_MIN;
          }

    out_pwm[0].ledbit = (1<<LED_RED1);
    out_pwm[1].ledbit = (1<<LED_GREEN1);

    #if (NUM_OF_SIGNALS == 2)
    out_pwm[2].ledbit = (1<<LED_RED2);
    out_pwm[3].ledbit = (1<<LED_GREEN2);
    #endif

    MyOpMode  = eeprom_read_byte(&EE_myOpMode);

    CurrentTarget = 0;
    CurrentAspect = eeprom_read_byte(&EE_LastState);

    set_new_light_val(CurrentAspect, 0xFF);

    if (eeprom_read_byte(&EE_myAdrH) & 0x80)
      {
        set_target(0,0); 
        out_pwm[0].rest    = 2*PROG_BLINK/2;
        out_pwm[0].offtime = PROG_BLINK/2;
        out_pwm[0].ontime  = PROG_BLINK/2;
      }  

    Communicate = 0; 
    Recstate = 1<<RECSTAT_WF_PREAMBLE;  

    sei();              // Global enable interrupts

    while(1)
      {
        cli();
        if (Communicate & (1<<C_Received)  )
          {
            sei();
            if (analyze_message() == 2)     // MyAdr empfangen
              {
                action();
              }
            Communicate &= ~(1<<C_Received); 
          }
        sei();
        sei();

        cli();
        if (Communicate & (1<<C_DoSave) )
          {
            sei();
            Communicate &= ~(1<<C_DoSave);
            eeprom_write_byte(&EE_LastState, CurrentAspect);   
          } 
        sei();
        sei();
        cli();
        if (Communicate & (1<<C_Dimmstep) )
          {
            sei();
            Communicate &= ~(1<<C_Dimmstep);
            dimmer();
          }

        if (PROG_PRESSED) DoProgramming();         
      }
  }


//=========================================================================
//
// Simulationen:
//
// 0: keine Simulation
// 1: Test der Empfangrootine; hierzu ein DCC Generator (aus opendcc/dccout.c)
// 2: Test der Portengine
//
#if (SIMULATION == 1)

//                                        Adressbyte   Datenbyte 
//                                         10AAAAAA    1aaaSCCR    // note: aaa = ~AAA
unsigned char message_adr001_out0_g[] = {0b10000001, 0b11111001};
unsigned char message_adr005_out1_g[] = {0b10000101, 0b11111011};
unsigned char message_adr380_out0_g[] = {0b10111100, 0b10101001};  // 380 = 0x17C = 0b101111100


// upstream interface:
struct
  {
    unsigned char size;
    unsigned char dcc[5];
  } next_message;

volatile unsigned char next_message_count;

enum do_states
  {                            // actual state
     dos_idle,
     dos_send_preamble,
     dos_send_bstart,
     dos_send_byte,
     dos_send_xor
  };

struct
  {
    enum do_states state;

    unsigned char ibyte;                            // current index of byte in message
    unsigned char bits_in_state;                    // Bits to output in this state
    unsigned char cur_byte;                         // current byte
    unsigned char xor_byte;                              // actual xor
    unsigned char current_dcc[5];                  // current message in output processing
    unsigned char bytes_in_message;                 // current size of message (decremented)
    unsigned char phase;
  } doi;


unsigned char dccbit;

void do_send(unsigned char mydccbit)
  {
    dccbit = mydccbit;
  }

void dcc_bit_generator(void)
  {
    switch (doi.state)
      {
        case dos_idle:
            do_send(1);
            if (next_message_count > 0)
              {
                memcpy(doi.current_dcc, next_message.dcc, sizeof(doi.current_dcc));
                doi.bytes_in_message = next_message.size;
                // no size checking - if (doi.cur_size > 5) doi.cur_size = 5;
                next_message_count--;
                doi.ibyte = 0;
                doi.xor_byte = 0;
                doi.bits_in_state = 14;
                doi.state = dos_send_preamble;
              }
            break;

        case dos_send_preamble:
            do_send(1);
            doi.bits_in_state--;
            if (doi.bits_in_state == 0)
                 doi.state = dos_send_bstart;
            break;

        case dos_send_bstart:
            do_send(0);
            if (doi.bytes_in_message == 0)
              { // message done, goto xor
                doi.cur_byte = doi.xor_byte;
                doi.state = dos_send_xor;
                doi.bits_in_state = 8;
              }
            else
              { // get next addr or data
                doi.bytes_in_message--;
                doi.cur_byte = doi.current_dcc[doi.ibyte++];
                doi.xor_byte ^= doi.cur_byte;
                doi.state = dos_send_byte;
                doi.bits_in_state = 8;
              }
            break;

        case dos_send_byte:
            if (doi.cur_byte & 0x80) do_send(1);
            else                    do_send(0);
            doi.cur_byte <<= 1;
            doi.bits_in_state--;
            if (doi.bits_in_state == 0)
              {
                doi.state = dos_send_bstart;
              }
            break;

        case dos_send_xor:
            if (doi.cur_byte & 0x80) do_send(1);
            else                    do_send(0);
            doi.cur_byte <<= 1;
            doi.bits_in_state--;
            if (doi.bits_in_state == 0)
              {
                doi.state = dos_idle;
              }
            break;
     }
  }

void dcc_generate_init(void)
  {
    doi.state = dos_idle;
  }


void simulat_receive(void)
  {
    dcc_generate_init();
    dcc_bit_generator();
    dcc_bit_generator();
    memcpy(next_message.dcc, message_adr001_out0_g, sizeof(doi.current_dcc));
    next_message.size = 2;
    next_message_count = 2;
    while(1)
      {
        dcc_bit_generator();
        dcc_receive();

        if (Communicate & (1<<C_Received)  )
          {
            if (analyze_message() == 2)     // MyAdr empfangen
              {
                action();
              }
            Communicate &= ~(1<<C_Received); 
          }
        if (Communicate & (1<<C_DoSave) )
          {
            Communicate &= ~(1<<C_DoSave);
            eeprom_write_byte(&EE_LastState, PORTB);   
          } 
        // if (PROG_PRESSED) DoProgramming();         
      }
 
  }


#endif // (SIMULATION == 1)


#if (SIMULATION == 2)
void simulat_mode_0_command_0(void)
{
  ReceivedCommand = 0x08; // Bit 3: active coil
  action();

      while(1)
      {
        sei();
        sei();
        cli();
        if (Communicate & (1<<C_Dimmstep) )
          {
            sei();
            Communicate &= ~(1<<C_Dimmstep);
            dimmer();
          }
      }

}


#endif // (SIMULATION == 2)

#if (SIMULATION == 3)
void simulat_mode_6_command_1(void)
{
  MyOpMode = 6;
  ReceivedCommand = 0b00001001; // Bit 3: active coil
  action();

      while(1)
      {
        sei();
        sei();
        cli();
        if (Communicate & (1<<C_Dimmstep) )
          {
            sei();
            Communicate &= ~(1<<C_Dimmstep);
            dimmer();
          }
      }

}


#endif // (SIMULATION == 3)
